import fakeredis._msgs as msgs
from fakeredis._command_args_parsing import extract_args
from fakeredis._commands import Key, command, StreamRangeTest
from fakeredis._helpers import SimpleError
from fakeredis._stream import XStream


class StreamsCommandsMixin:
    @command(name="XADD", fixed=(Key(),), repeat=(bytes,), )
    def xadd(self, key, *args):
        # TODO: handle id_str not *, MAXLEN, MINID, LIMIT
        (nomkstream, limit,), left_args = extract_args(args, ('nomkstream', '+limit'), error_on_unexpected=False)
        if nomkstream and key.value is None:
            return None
        id_str = left_args[0]
        elements = left_args[1:]
        if not elements or len(elements) % 2 != 0:
            raise SimpleError(msgs.WRONG_ARGS_MSG6.format('XADD'))
        stream = key.value or XStream()
        id_str = stream.add(elements, id_str=id_str)
        if id_str is not None:
            key.update(stream)
        return id_str

    @command(name="XLEN", fixed=(Key(XStream),))
    def xlen(self, key):
        if key.value is None:
            return 0
        return len(key.value)

    def _xrange(self, key, _min, _max, reverse, count, ):
        if key.value is None:
            return None
        if count is None:
            count = len(key.value)
        res = key.value.irange(
            _min.value, _max.value,
            exclusive=(_min.exclusive, _max.exclusive),
            reverse=reverse)
        return res[:count]

    @command(name="XRANGE", fixed=(Key(XStream), StreamRangeTest, StreamRangeTest), repeat=(bytes,))
    def xrange(self, key, _min, _max, *args):
        (count,), _ = extract_args(args, ('+count',))
        return self._xrange(key, _min, _max, False, count)

    @command(name="XREVRANGE", fixed=(Key(XStream), StreamRangeTest, StreamRangeTest), repeat=(bytes,))
    def xrevrange(self, key, _min, _max, *args):
        (count,), _ = extract_args(args, ('+count',))
        return self._xrange(key, _max, _min, True, count)



import bisect
import time
from typing import List, Union, Tuple

from fakeredis._commands import StreamRangeTest


class XStream:
    def __init__(self):
        # Values:
        # [
        #    ((timestamp,sequence), [field1, value1, field2, value2, ...])
        #    ((timestamp,sequence), [field1, value1, field2, value2, ...])
        # ]
        self._values = list()

    def add(self, fields: List, id_str: str = '*') -> Union[None, bytes]:
        assert len(fields) % 2 == 0
        if id_str != b'*':
            ts_seq = StreamRangeTest.parse_id(id_str)
        else:
            ts, seq = int(time.time() + 1), 1
            if (len(self._values) > 0
                    and self._values[-1][0][0] == ts
                    and self._values[-1][0][1] >= seq):
                seq = self._values[-1][0][1] + 1
            ts_seq = (ts, seq)

        if len(self._values) > 0 and self._values[-1][0] > ts_seq:
            return None
        new_val = (ts_seq, list(fields))
        self._values.append(new_val)
        return f'{ts_seq[0]}-{ts_seq[1]}'.encode()

    def __len__(self):
        return len(self._values)

    def __iter__(self):
        def gen():
            for val in self._values:
                ts_seq = val[0]
                yield f"{ts_seq[0]}-{ts_seq[1]}"

        return gen()

    def find_index(self, id_str: str):
        ts_seq = StreamRangeTest.parse_id(id_str)
        return bisect.bisect_left(self._values, ts_seq, key=lambda x: x[0])

    def _format_record(self, record):
        return [f'{record[0][0]}-{record[0][1]}'.encode(), list(record[1:])]

    def irange(self,
               start, stop,
               exclusive: Tuple[bool, bool] = (True, True),
               reverse=False):
        def match(record):
            res = stop > record[0] > start
            res = res or (not exclusive[0] and record[0] == start)
            res = res or (not exclusive[1] and record[0] == stop)
            return res

        matches = filter(match, self._values)
        res = list()
        for record in matches:
            res.append(self._format_record(record))
        if reverse:
            return list(reversed(res))
        return res


if __name__ == '__main__':
    stream = XStream()
    stream.add([0, 0, 1, 1, 2, 2, 3, 3], '0-1')
    stream.add([1, 1, 2, 2, 3, 3, 4, 4], '1-2')
    stream.add([2, 2, 3, 3, 4, 4], '1-3')
    stream.add([3, 3, 4, 4], '2-1')
    stream.add([3, 3, 4, 4], '2-2')
    stream.add([3, 3, 4, 4], '3-1')
    for i in stream:
        print(i)
    assert stream.find_index('1-2') == 1
    assert stream.find_index('0-1') == 0
    assert stream.find_index('2-1') == 3
    assert stream.find_index('1-4') == 3
    lst = stream.irange('0-2', '3-0')
    for i in lst:
        print(i)
